########### R function: SolveTwoLevelSparseLeastSquares ###########

# For solving a two-level sparse matrix least squares problem
# according to Algorithm 2 of Nolan, Menictas & Wand (2018).

# Last changed: 29 MAY 2019

SolveTwoLevelSparseLeastSquares <- function(b,B,Bdot)
{
   # Extract dimension variables:

   m <- length(b) ; p <- ncol(B[[1]]) ; q <- ncol(Bdot[[1]])

   if (length(B)!=m) stop("B is illegal.\n")
   if (length(Bdot)!=m) stop("Bdot is illegal.\n")

   # Obtain the x1 and Asup11 solutions:

   omega3 <- NULL ; Omega4 <- NULL  ; Rlist <- vector("list",m)
   c1 <- vector("list",m) ; C1 <- vector("list",m)
   for (i in 1:m)
   {
      qrObj <- qr(Bdot[[i]]) ; Rlist[[i]] <- qr.R(qrObj)
      c0i <- qr.qty(qrObj,b[[i]])  ;  C0i <- qr.qty(qrObj,B[[i]])
      c1[[i]] <- c0i[1:q]  ;  c2i <- c0i[-(1:q)]  ;  omega3 <- c(omega3, c2i)
      C1[[i]] <- C0i[1:q,]  
      C2i <- C0i[-(1:q),]  ;  Omega4 <- rbind(Omega4, C2i)
      if (p==1) C1[[i]] <- as.matrix(C1[[i]])
      if (q==1) C1[[i]] <- t(as.matrix(C1[[i]]))
   }
   qrObj <- qr(Omega4)  ;  R <- qr.R(qrObj)
   cvec <- qr.qty(qrObj,omega3)[1:p]  ;  x1 <- backsolve(R,cvec)
   Asup11 <- backsolve(R,t(backsolve(R,diag(p))))

   # Obtain the x2, Asup12 and Asup22 solutions:

   x2 <- vector("list",m)
   Asup12 <- vector("list",m)  ;  Asup22 <- vector("list",m)
   for (i in 1:m)
   {
      x2[[i]] <- backsolve(Rlist[[i]],(c1[[i]] - crossprod(t(C1[[i]]),x1))) 
      Asup12[[i]] <- -tcrossprod(Asup11,backsolve(Rlist[[i]],C1[[i]]))
      Asup22[[i]] <- backsolve(Rlist[[i]],(t(backsolve(Rlist[[i]],diag(q)))
                                           - crossprod(t(C1[[i]]),Asup12[[i]])))
   }

   # Obtain the log|A| value:
   
   logDetA <- 2*sum(log(abs(diag(R))))
   for (i in 1:m) 
      logDetA <- logDetA + 2*sum(log(abs(diag(Rlist[[i]]))))

   # Return output list:

   return(list(x1=x1,Asup11=Asup11,x2=x2,Asup22=Asup22,Asup12=Asup12,logDetA=logDetA))
}

########### End of SolveTwoLevelSparseLeastSquares ############
