########### R function: threeLevGrpSpecCurLogML ###########

# For computing the marginal log-likelihood for mean field variational
# Bayesian inference for the Bayesian three-level group specific curve model 
# of Menictas, Nolan, Simpson and Wand (2020).

# Last changed: 24 FEB 2020

threeLevGrpSpecCurvLogML <- function(
    p, m, nVec, oMat, q1, q2, Kgbl, Kgrp1, Kgrp2, mu.beta, Sigma.beta, yList, CgblList, Cgrp1List, Cgrp2List,
    mu.q.beta, Sigma.q.beta, mu.q.betauGbl, Sigma.q.betauGbl, mu.q.uLinGrp1, Sigma.q.uLinGrp1, mu.q.uLinGrp2,
    Sigma.q.uLinGrp2, E.q.betauGblLinGrp1, E.q.betauGblLinGrp2, E.q.uLinGrp1uLinGrp2, M.q.inv.SigmaL1,
    M.q.inv.SigmaL2, mu.q.recip.sigsq.gbl, mu.q.recip.sigsq.grp1, mu.q.recip.sigsq.grp2,
    nu.eps, nu.gbl, nu.grp1, nu.grp2, nu.SigmaL1, nu.SigmaL2, mu.q.recip.sigsq.eps, mu.q.recip.a.eps,
    xi.q.sigsq.eps, lambda.q.sigsq.eps, s.eps, xi.q.a.sigsq.eps,lambda.q.a.sigsq.eps, mu.q.recip.a.gbl,
    xi.q.sigsq.gbl, lambda.q.sigsq.gbl, s.gbl, xi.q.a.sigsq.gbl, lambda.q.a.sigsq.gbl, mu.q.recip.a.grp1,
    xi.q.sigsq.grp1, lambda.q.sigsq.grp1, s.grp1, xi.q.a.sigsq.grp1, lambda.q.a.sigsq.grp1, mu.q.recip.a.grp2,
    xi.q.sigsq.grp2, lambda.q.sigsq.grp2, s.grp2, xi.q.a.sigsq.grp2, lambda.q.a.sigsq.grp2, M.q.inv.A.SigmaL1,
    Lambda.q.SigmaL1, M.q.inv.A.SigmaL2, Lambda.q.SigmaL2, xi.q.A.SigmaL1, xi.q.SigmaL1, s.SigmaL1,
    Lambda.q.A.SigmaL1, xi.q.A.SigmaL2, xi.q.SigmaL2, s.SigmaL2, Lambda.q.A.SigmaL2, det.Sigma.q.betau
  ) 
{

    logMlVal <- - 0.5 * (sum(oMat) + 4) * log(pi)
    logMlVal <- logMlVal + (0.5 * (p + Kgbl + m * (q1 + Kgrp1) + sum(nVec) * (q2 + Kgrp2)))
    logMlVal <- logMlVal - 0.5 * (sum(oMat) + nu.eps + nu.gbl + nu.grp1 + nu.grp2) * log(2)
    logMlVal <- logMlVal + (p + m + sum(nVec) + nu.SigmaL1 + nu.SigmaL2) * log(2)
    logMlVal <- logMlVal - 0.5 * sum(diag(solve(Sigma.beta) %*% (tcrossprod(mu.q.beta - mu.beta ) + Sigma.q.beta)))
    logMlVal <- logMlVal - 0.5 * as.numeric( determinant( Sigma.beta )$modulus)
    logMlVal <- logMlVal + 0.5 * det.Sigma.q.betau

    LambdaVal1 <- matrix(0,q1,q1)
    LambdaVal2 <- matrix(0,q2,q2)
    lambdaVal1 <- 0 ; lambdaVal2 <- 0 ; lambdaEpsVal <- 0
    for (i in 1:m) {
       LambdaVal1 <- LambdaVal1 + tcrossprod(mu.q.uLinGrp1[[i]][1:q1]) + Sigma.q.uLinGrp1[[i]][1:q1,1:q1]
       lambdaVal1 <- lambdaVal1 + (sum(mu.q.uLinGrp1[[i]][(p+1):(p+Kgrp1)]^2)
                       + sum(diag(Sigma.q.uLinGrp1[[i]][(p+1):(p+Kgrp1),(p+1):(p+Kgrp1)])))
       for (j in 1:nVec[i])
       {
          LambdaVal2 <- LambdaVal2 + tcrossprod(mu.q.uLinGrp2[[i]][[j]][1:q2]) + Sigma.q.uLinGrp2[[i]][[j]][1:q2,1:q2]
          lambdaVal2 <- lambdaVal2 + (sum(mu.q.uLinGrp2[[i]][[j]][(p+1):(p+Kgrp2)]^2)
                         + sum(diag(Sigma.q.uLinGrp2[[i]][[j]][(p+1):(p+Kgrp2), (p+1):(p+Kgrp2)])))

          residVec <- (yList[[i]][[j]] - as.vector(CgblList[[i]][[j]]%*%mu.q.betauGbl)
                      - as.vector(Cgrp1List[[i]][[j]]%*%mu.q.uLinGrp1[[i]])
                      - as.vector(Cgrp2List[[i]][[j]]%*%mu.q.uLinGrp2[[i]][[j]]))

          lambdaEpsVal <- lambdaEpsVal + sum(residVec^2)
          lambdaEpsVal <- lambdaEpsVal + sum(diag(tcrossprod(crossprod(CgblList[[i]][[j]]),Sigma.q.betauGbl)))
          lambdaEpsVal <- lambdaEpsVal + sum(diag(tcrossprod(crossprod(Cgrp1List[[i]][[j]]),Sigma.q.uLinGrp1[[i]])))
          lambdaEpsVal <- lambdaEpsVal + sum(diag(tcrossprod(crossprod(Cgrp2List[[i]][[j]]),Sigma.q.uLinGrp2[[i]][[j]])))
          lambdaEpsVal <- lambdaEpsVal + 2*sum(diag(tcrossprod(crossprod(Cgrp1List[[i]][[j]],CgblList[[i]][[j]]),
                                               t(E.q.betauGblLinGrp1[[i]]))))
          lambdaEpsVal <- lambdaEpsVal + 2*sum(diag(tcrossprod(crossprod(Cgrp2List[[i]][[j]],CgblList[[i]][[j]]),
                                               t(E.q.betauGblLinGrp2[[i]][[j]]))))
          lambdaEpsVal <- lambdaEpsVal + 2*sum(diag(tcrossprod(crossprod(Cgrp1List[[i]][[j]],Cgrp2List[[i]][[j]]),
                                               E.q.uLinGrp1uLinGrp2[[i]][[j]])))
       }
    }

    logMlVal <- logMlVal - 0.5 * sum(diag(M.q.inv.SigmaL1 %*% LambdaVal1))
    logMlVal <- logMlVal - 0.5 * sum(diag(M.q.inv.SigmaL2 %*% LambdaVal2))
    logMlVal <- logMlVal - 0.5 * mu.q.recip.sigsq.gbl %*% (sum(mu.q.betauGbl[(p+1):(p+Kgbl)]^2)
                 + sum(diag(Sigma.q.betauGbl[(p+1):(p+Kgbl), (p+1):(p+Kgbl)])))
    logMlVal <- logMlVal - 0.5 * mu.q.recip.sigsq.grp1 %*% lambdaVal1
    logMlVal <- logMlVal - 0.5 * mu.q.recip.sigsq.grp2 %*% lambdaVal2

    logMlVal <- logMlVal - lgamma( nu.eps / 2 )
    logMlVal <- logMlVal - 0.5 * mu.q.recip.sigsq.eps * mu.q.recip.a.eps
    logMlVal <- logMlVal - 0.5 * xi.q.sigsq.eps * log( lambda.q.sigsq.eps / 2 )
    logMlVal <- logMlVal + lgamma( xi.q.sigsq.eps / 2 )
    logMlVal <- logMlVal + 0.5 * lambda.q.sigsq.eps * mu.q.recip.sigsq.eps
    logMlVal <- logMlVal - 0.5 * log( 2 * nu.eps * s.eps^2 )
    logMlVal <- logMlVal - (1/(2 * nu.eps * s.eps^2)) * mu.q.recip.a.eps
    logMlVal <- logMlVal - 0.5 * xi.q.a.sigsq.eps * log( lambda.q.a.sigsq.eps / 2 )
    logMlVal <- logMlVal + lgamma( xi.q.a.sigsq.eps / 2 )
    logMlVal <- logMlVal + 0.5 * lambda.q.a.sigsq.eps * mu.q.recip.a.eps

    logMlVal <- logMlVal - lgamma( nu.gbl / 2 )
    logMlVal <- logMlVal - 0.5 * mu.q.recip.sigsq.gbl * mu.q.recip.a.gbl
    logMlVal <- logMlVal - 0.5 * xi.q.sigsq.gbl * log( lambda.q.sigsq.gbl / 2)
    logMlVal <- logMlVal + lgamma( xi.q.sigsq.gbl / 2 )
    logMlVal <- logMlVal + 0.5 * lambda.q.sigsq.gbl * mu.q.recip.sigsq.gbl
    logMlVal <- logMlVal - 0.5 * log( 2 * nu.gbl * s.gbl^2 )
    logMlVal <- logMlVal - (1/(2 * nu.gbl * s.gbl^2)) * mu.q.recip.a.gbl
    logMlVal <- logMlVal - 0.5 * xi.q.a.sigsq.gbl * log( lambda.q.a.sigsq.gbl / 2)
    logMlVal <- logMlVal + lgamma( xi.q.a.sigsq.gbl / 2 )
    logMlVal <- logMlVal + 0.5 * lambda.q.a.sigsq.gbl * mu.q.recip.a.gbl

    logMlVal <- logMlVal - lgamma( nu.grp1 / 2 )
    logMlVal <- logMlVal - 0.5 * mu.q.recip.sigsq.grp1 * mu.q.recip.a.grp1
    logMlVal <- logMlVal - ( 0.5 * xi.q.sigsq.grp1 * log( lambda.q.sigsq.grp1 / 2) )
    logMlVal <- logMlVal + lgamma( xi.q.sigsq.grp1 / 2 )
    logMlVal <- logMlVal + 0.5 * lambda.q.sigsq.grp1 * mu.q.recip.sigsq.grp1
    logMlVal <- logMlVal - 0.5 * log( 2 * nu.grp1 * s.grp1^2 )
    logMlVal <- logMlVal - (1/(2 * nu.grp1 * s.grp1^2)) * mu.q.recip.a.grp1
    logMlVal <- logMlVal - ( 0.5 * xi.q.a.sigsq.grp1 * log( lambda.q.a.sigsq.grp1 / 2) )
    logMlVal <- logMlVal + lgamma( xi.q.a.sigsq.grp1 / 2 )
    logMlVal <- logMlVal + 0.5 * lambda.q.a.sigsq.grp1 * mu.q.recip.a.grp1

    logMlVal <- logMlVal - lgamma( nu.grp2 / 2 )
    logMlVal <- logMlVal - 0.5 * mu.q.recip.sigsq.grp2 * mu.q.recip.a.grp2
    logMlVal <- logMlVal - ( 0.5 * xi.q.sigsq.grp2 * log( lambda.q.sigsq.grp2 / 2) )
    logMlVal <- logMlVal + lgamma( xi.q.sigsq.grp2 / 2 )
    logMlVal <- logMlVal + 0.5 * lambda.q.sigsq.grp2 * mu.q.recip.sigsq.grp2
    logMlVal <- logMlVal - 0.5 * log( 2 * nu.grp2 * s.grp2^2 )
    logMlVal <- logMlVal - (1/(2 * nu.grp2 * s.grp2^2)) * mu.q.recip.a.grp2
    logMlVal <- logMlVal - ( 0.5 * xi.q.a.sigsq.grp2 * log( lambda.q.a.sigsq.grp2 / 2) )
    logMlVal <- logMlVal + lgamma( xi.q.a.sigsq.grp2 / 2 )
    logMlVal <- logMlVal + 0.5 * lambda.q.a.sigsq.grp2 * mu.q.recip.a.grp2

    logMlVal <- logMlVal - 0.5 * sum( diag( M.q.inv.A.SigmaL1 %*% M.q.inv.SigmaL1 ) )
    logMlVal <- logMlVal + 0.5 * sum( diag( Lambda.q.SigmaL1 %*% M.q.inv.SigmaL1 ) )

    logMlVal <- logMlVal - 0.5 * sum( diag( M.q.inv.A.SigmaL2 %*% M.q.inv.SigmaL2 ) )
    logMlVal <- logMlVal + 0.5 * sum( diag( Lambda.q.SigmaL2 %*% M.q.inv.SigmaL2 ) )

    logMlVal <- logMlVal + sum( lgamma( 0.5 * ( xi.q.A.SigmaL1 + 2 - (1:q1) ) ) )
    logMlVal <- logMlVal - sum( lgamma( 0.5 * ( nu.SigmaL1 + 2 * q1 - (1:q1) ) ) )
    logMlVal <- logMlVal - 0.5 * ( xi.q.SigmaL1 - q1 + 1 ) * as.numeric( determinant( Lambda.q.SigmaL1 )$modulus )
    logMlVal <- logMlVal + sum( lgamma( 0.5 * ( xi.q.SigmaL1 + 2 - (1:q1) ) ) )
    logMlVal <- logMlVal - 0.5 * sum( (1/(nu.SigmaL1 * s.SigmaL1)) * diag(M.q.inv.A.SigmaL1) )
    logMlVal <- logMlVal - sum( lgamma( 0.5 * ( 3 - (1:q1) ) ) )
    logMlVal <- logMlVal - 0.5 * ( xi.q.A.SigmaL1 - q1 + 1 ) * as.numeric( determinant( Lambda.q.A.SigmaL1 )$modulus )
    logMlVal <- logMlVal + sum(diag( Lambda.q.A.SigmaL1 %*% M.q.inv.A.SigmaL1 ))

    logMlVal <- logMlVal + sum( lgamma( 0.5 * ( xi.q.A.SigmaL2 + 2 - (1:q2) ) ) )
    logMlVal <- logMlVal - sum( lgamma( 0.5 * ( nu.SigmaL2 + 2 * q2 - (1:q2) ) ) )
    logMlVal <- logMlVal - 0.5 * ( xi.q.SigmaL2 - q2 + 1 ) * as.numeric( determinant( Lambda.q.SigmaL2 )$modulus )
    logMlVal <- logMlVal + sum( lgamma( 0.5 * ( xi.q.SigmaL2 + 2 - (1:q2) ) ) )
    logMlVal <- logMlVal - 0.5 * sum( (1/(nu.SigmaL2 * s.SigmaL2)) * diag(M.q.inv.A.SigmaL2) )
    logMlVal <- logMlVal - sum( lgamma( 0.5 * ( 3 - (1:q2) ) ) )
    logMlVal <- logMlVal - 0.5 * ( xi.q.A.SigmaL2 - q2 + 1 ) * as.numeric( determinant( Lambda.q.A.SigmaL2 )$modulus )
    logMlVal <- logMlVal + sum(diag( Lambda.q.A.SigmaL2 %*% M.q.inv.A.SigmaL2 ))

    logMlVal <- logMlVal - 0.5 * mu.q.recip.sigsq.eps * lambdaEpsVal

    return(logMlVal)
}

########### End of threeLevGrpSpecCurLogML ############
